from io import StringIO
from typing import List, Dict, Tuple, Literal
import babel.numbers
import matplotlib.pyplot as plt
import babel

_COLOR_MAP_AREA_TYPES = {
    'Bebygd': '#fcdbd6',
    'Skog': '#9ecc73',
    'Åpen fastmark': '#d9d9d9',
    'Samferdsel': '#b3784c',
    'Fulldyrka jord': '#ffd16e',
    'Ferskvann': '#91e7ff',
    'Hav': '#d2ffff',
    'Ikke kartlagt': '#cccccc',
    'Innmarksbeite': '#fffA56',
    'Myr': '#73dfe1',
    'Overflatedyrka jord': '#ffcd56',
    'Snøisbre': '#ffffff'
}

_COLOR_MAP_BUILDINGS = {
    'Bolig': '#c48723',
    'Fritidsbolig - hytte': '#dcaa27',
    'Industri og lagerbygning': '#74a3d4',
    'Kontor- og forretningsbygning': '#74a3d4',
    'Samferdsels- og kommunikasjonsbygning': '#74a3d4',
    'Hotell og restaurantbygning': '#74a3d4',
    'Skole-, kultur-, idrett-, forskningsbygning': '#74a3d4',
    'Helse- og omsorgsbygning': '#74a3d4',
    'Fengsel, beredskapsbygning, mv.': '#74a3d4'
}


def create_pie_chart(fact_list: List[Dict], type: Literal['AREA_TYPES', 'BUILDINGS']) -> Tuple[str, List[Dict]]:
    if type == 'AREA_TYPES':
        values, values_formatted, labels, legend, colors = _get_options_for_area_types(
            fact_list)
    else:
        values, values_formatted, labels, legend, colors = _get_options_for_buildings(
            fact_list)

    fig, ax = plt.subplots(figsize=(5, 5))

    fig.tight_layout(pad=0)
    fig.set_frameon(False)

    ax.axis('off')
    ax.pie(values, colors=colors, wedgeprops={
           'linewidth': 0.5, 'edgecolor': 'white'})

    string = StringIO()
    fig.savefig(string, bbox_inches='tight', pad_inches=-0.4, format='svg')

    options: List[Dict] = []

    for i in range(len(values)):
        options.append({
            'value': values[i],
            'value_formatted': values_formatted[i],
            'label': labels[i],
            'legend_item': legend[i],
            'color': colors[i]
        })

    return string.getvalue(), options


def _get_options_for_area_types(fact_list: List[Dict]) -> Tuple[List[int], List[str], List[str], List[str], List[str]]:
    area_types = fact_list[0]['data']['areaTypes']

    sorted_area_types = sorted(
        area_types, key=lambda area_type: area_type['area'], reverse=True)

    values: List[int] = []
    values_formatted: List[str] = []
    labels: List[str] = []
    legend: List[str] = []
    colors: List[str] = []

    for area_type in sorted_area_types:
        value = round(area_type['area'])

        if value == 0:
            continue

        label = area_type['areaType']
        value_formatted = babel.numbers.format_decimal(value, locale='nb_NO')

        values.append(value)
        values_formatted.append(value_formatted)
        labels.append(label)
        legend.append(f'{label}:  {value_formatted} m²')
        colors.append(_COLOR_MAP_AREA_TYPES[label])

    return values, values_formatted, labels, legend, colors


def _get_options_for_buildings(fact_list: List[Dict]) -> Tuple[List[int], List[str], List[str], List[str], List[str]]:
    buildings = fact_list[1]['data']
    sorted_buildings = sorted(
        buildings, key=lambda building: building['count'], reverse=True)

    values: List[int] = []
    values_formatted: List[str] = []
    labels: List[str] = []
    legend: List[str] = []
    colors: List[str] = []

    for building in sorted_buildings:
        value = round(building['count'])

        if value == 0:
            continue

        label = building['category']
        value_formatted = babel.numbers.format_decimal(value, locale='nb_NO')
        
        values.append(value)
        values_formatted.append(value_formatted)
        labels.append(label)
        legend.append(f'{label}:  {value_formatted} stk.')
        colors.append(_COLOR_MAP_BUILDINGS[label])

    return values, values_formatted, labels, legend, colors


__all__ = ['create_pie_chart']



from typing import Dict, List, Any
from osgeo import ogr, osr
from ..utils.helpers.geometry import create_feature_collection, transform_geometry


def query_gdal(driver_name: str, data_source: Any, filter: str, geometry: ogr.Geometry, epsg: int) -> Dict:
    driver: ogr.Driver = ogr.GetDriverByName(driver_name)
    ds: ogr.DataSource = driver.Open(data_source)
    layer: ogr.Layer = ds.GetLayer(0)

    sr: osr.SpatialReference = layer.GetSpatialRef()
    auth_code: str = sr.GetAuthorityCode(None)
    target_epsg = int(auth_code)

    if target_epsg != epsg:
        input_geometry = transform_geometry(geometry, epsg, target_epsg)
    else:
        input_geometry = geometry

    layer.SetSpatialFilter(input_geometry)

    if filter:
        layer.SetAttributeFilter(filter)

    ogr_feature: ogr.Feature
    features: List[Dict] = []

    for ogr_feature in layer:
        feature_geom: ogr.Geometry = ogr_feature.GetGeometryRef()
        clone: ogr.Geometry = feature_geom.Clone()
        properties = {}
        field_count: int = ogr_feature.GetFieldCount()

        for i in range(field_count):
            field: ogr.FieldDefn = ogr_feature.GetFieldDefnRef(i)
            name = field.GetName()
            value = ogr_feature.GetField(i)
            properties[name] = value

        feature = {
            'geometry': clone,
            'properties': properties
        }

        features.append(feature)

    response = create_feature_collection(features, target_epsg)

    return response



import logging
import time
from io import BytesIO
from urllib.parse import urlparse, parse_qs
from typing import List, Dict, Tuple
from pebble import ThreadPool, ProcessPool
from concurrent.futures import TimeoutError, as_completed
from osgeo import ogr, osr
import pandas as pd
import geopandas as gpd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.image import AxesImage
import cartopy.crs as ccrs
import cartopy.io.ogc_clients as ogcc
from cartopy.mpl.geoaxes import GeoAxes
import contextily as ctx
from shapely import box, Polygon
from ..models.analysis import Analysis
from ..models.fact_sheet import FactSheet

ogcc.METERS_PER_UNIT['EPSG:3857'] = 1
ogcc._URN_TO_CRS['EPSG:3857'] = ccrs.GOOGLE_MERCATOR

_LOGGER = logging.getLogger(__name__)
_WMTS_URL = 'https://cache.kartverket.no/v1/wmts/1.0.0/WMTSCapabilities.xml?request=GetCapabilities'
_DPI = 100
_TIMEOUT_SECONDS = 120


def generate_map_images(analyses: List[Analysis], fact_sheet: FactSheet | None) -> List[Tuple[str, str, bytes | None]]:
    params: List[Dict] = []
    params.extend(_get_params_for_analyses(analyses))

    if fact_sheet:
        params.append(_get_params_for_fact_sheet(fact_sheet))

    start = time.time()
    results: List[Tuple[str, bytes | None]] = []

    with ProcessPool(max_workers=8) as pool:
        futures = [pool.submit(
            _generate_map_image, _TIMEOUT_SECONDS, **kwargs) for kwargs in params]

        for future in as_completed(futures):
            try:
                results.append(future.result())
            except TimeoutError as error:
                _LOGGER.error('Map image generation timed out')
            except Exception as error:
                _LOGGER.error('Map image generation failed: %s' % error)

    # autopep8: off
    _LOGGER.info(f'Generated {len(results)} map images in {round(time.time() - start, 2)} sec.')
    # autopep8: on

    return results


def _generate_map_image(**kwargs) -> Tuple[str, str, bytes | None]:
    id: str = kwargs['id']
    name: str = kwargs['name']
    wkt_str: str = kwargs['geometry']
    gdf = gpd.GeoSeries.from_wkt([wkt_str])
    crs_epsg = ccrs.epsg('3857')

    size: Tuple[int, int] = kwargs.get('size', (800, 600))
    width, height = size
    figsize = _get_figsize(width, height)

    fig, ax = plt.subplots(figsize=figsize, dpi=_DPI, subplot_kw={
        'projection': crs_epsg})

    ax.axis('off')

    gdf.plot(ax=ax, edgecolor='#d33333',
             facecolor='none', linewidth=3)

    buffer: str = kwargs.get('buffer')

    if buffer:
        buffer_row = gpd.GeoSeries.from_wkt([buffer])
        gdf: gpd.GeoSeries = pd.concat([gdf, buffer_row], ignore_index=True)
        gdf.plot(ax=ax, edgecolor='#d33333', facecolor='none',
                 linestyle='--', linewidth=1.5)

    ext_bbox = _extend_bbox_to_aspect_ratio(gdf.total_bounds, width/height)
    ext_bbox_row = gpd.GeoSeries.from_wkt([ext_bbox.wkt])

    gdf: gpd.GeoSeries = pd.concat([gdf, ext_bbox_row], ignore_index=True)
    gdf.plot(ax=ax, facecolor='none', linewidth=0)

    grayscale: bool = kwargs.get('grayscale', False)
    use_wmts: bool = kwargs.get('use_wmts', True)

    try:
        _add_basemap(ax, use_wmts, grayscale)
    except:
        return id, name, None

    wms: Dict = kwargs.get('wms')

    if wms:
        try:
            _add_wms(ax, wms.get('url'), wms.get('layers'))
        except:
            return id, name, None

    image_bytes = _convert_to_bytes(fig)

    plt.close()

    return id, name, image_bytes


def _get_params_for_analyses(analyses: List[Analysis]) -> List[Dict]:
    params: List[Dict] = []

    for analysis in analyses:
        url, layers = _parse_wms_url(analysis.raster_result_map)

        if analysis.buffer > 0:
            buffered_geom: ogr.Geometry = analysis.geometry.Buffer(analysis.buffer)
            buffer = _get_wkt_str(buffered_geom)
        else:
            buffer = None
    
        params.append({
            'id': str(analysis.config_id),
            'name': analysis.config.name,
            'size': (1280, 720),
            'geometry': _get_wkt_str(analysis.geometry),
            'buffer': buffer,
            'grayscale': True,
            'wms': {
                'url': url,
                'layers': layers
            }
        })

    return params


def _get_params_for_fact_sheet(fact_sheet: FactSheet) -> Dict:
    if fact_sheet.buffer > 0:
        buffered_geom: ogr.Geometry = fact_sheet.geometry.Buffer(fact_sheet.buffer)
        buffer = _get_wkt_str(buffered_geom)
    else:
        buffer = None

    return {
        'id': 'omraade',
        'name': 'omraade',
        'size': (1280, 720),
        'geometry': _get_wkt_str(fact_sheet.geometry),
        'buffer': buffer,
        'grayscale': False        
    }


def _add_basemap(ax: GeoAxes, use_wmts: bool, grayscale: bool) -> None:
    if use_wmts:
        layer_name = 'topograatone' if grayscale else 'topo'
        ax.add_wmts(_WMTS_URL, layer_name=layer_name)
    else:
        basemap_src = ctx.providers.CartoDB.Positron if grayscale else ctx.providers.OpenStreetMap.Mapnik
        ctx.add_basemap(ax, source=basemap_src)


def _add_wms(ax: GeoAxes, url: str, layers: List[str]) -> AxesImage:
    return ax.add_wms(wms=url, layers=layers)


def _extend_bbox_to_aspect_ratio(bounds: Tuple[float, float, float, float], target_aspect_ratio: float) -> Polygon:
    minx, miny, maxx, maxy = bounds
    width = maxx - minx
    height = maxy - miny
    center_x = (minx + maxx) / 2
    center_y = (miny + maxy) / 2

    current_aspect = width / height

    if current_aspect > target_aspect_ratio:
        new_height = width / target_aspect_ratio
        new_width = width
    else:
        new_width = height * target_aspect_ratio
        new_height = height

    half_width = new_width / 2
    half_height = new_height / 2

    new_minx = center_x - half_width
    new_maxx = center_x + half_width
    new_miny = center_y - half_height
    new_maxy = center_y + half_height

    return box(new_minx, new_miny, new_maxx, new_maxy)


def _convert_to_bytes(fig: Figure) -> bytes:
    fig.tight_layout(pad=0)
    fig.set_frameon(False)

    byte_arr = BytesIO()
    fig.savefig(byte_arr, format='png', dpi=100)

    return byte_arr.getvalue()


def _get_wkt_str(geometry: ogr.Geometry) -> str:
    src_epsg = _get_epsg_code(geometry)
    coord_trans = _get_coordinate_transformation(src_epsg, 3857)
    clone: ogr.Geometry = geometry.Clone()
    
    clone.Transform(coord_trans)

    return clone.ExportToWkt()


def _get_epsg_code(geometry: ogr.Geometry) -> int:
    sr: osr.SpatialReference = geometry.GetSpatialReference()
    epsg_str: str = sr.GetAuthorityCode(None)

    return int(epsg_str)


def _get_coordinate_transformation(src_epsg: int, target_epsg: int) -> osr.CoordinateTransformation:
    source: osr.SpatialReference = osr.SpatialReference()
    source.ImportFromEPSG(src_epsg)

    target: osr.SpatialReference = osr.SpatialReference()
    target.ImportFromEPSG(target_epsg)

    return osr.CoordinateTransformation(source, target)


def _get_figsize(width: int, height: int) -> Tuple[int, int]:
    return (width / _DPI, height / _DPI)


def _parse_wms_url(url: str) -> Tuple[str, List[str]]:
    parsed = urlparse(url)
    query_strings = parse_qs(parsed.query)
    layers_list: List[str] = query_strings.get('layers', [])

    base_url = f'{parsed.scheme}://{parsed.netloc}{parsed.path}'
    layers = layers_list[0].split(',') if len(layers_list) == 1 else []

    return base_url, layers


__all__ = ['generate_map_images']



from datetime import datetime
import babel.numbers
import markdown as md


def markdown(value: str) -> str:
    if not value:
        return ''
    
    return md.markdown(value)


def format_number(value: int | float) -> str:
    if not value:
        return ''
    
    return babel.numbers.format_decimal(value, locale='nb_NO')


def format_datetime(date_str: str) -> str:
    if not date_str:
        return ''

    datetime_object = datetime.strptime(date_str, '%Y-%m-%dT%H:%M:%S')

    return datetime_object.strftime('%d.%m.%Y')



from typing import List, Dict
import babel.numbers


def get_results_by_result_status(response: Dict, result_status: str) -> List[Dict]:
    analyses = [
        analysis for analysis in response['resultList'] if analysis['resultStatus'] == result_status]

    if not analyses:
        return []

    if result_status == 'HIT_RED' or result_status == 'HIT_YELLOW':
        return sorted(analyses, key=lambda analysis: (analysis['hitArea'] or 0, analysis['themes'][0]), reverse=True)

    elif result_status == 'NO_HIT_GREEN':
        return sorted(analyses, key=lambda analysis: (analysis['distanceToObject'], analysis['themes'][0]), reverse=True)

    return analyses


def get_not_relevant_dataset_titles(response: Dict) -> List[str]:
    analyses = get_results_by_result_status(response, 'NOT-RELEVANT')

    if not analyses:
        return []

    titles: List[str] = [analysis.get('runOnDataset', {}).get(
        'title') or analysis['title'] for analysis in analyses]
        
    distinct = list(set(titles))
    distinct.sort()

    return distinct


def get_result_title(result: Dict) -> str:
    if 'runOnDataset' in result:
        dataset_title = f"«{result['runOnDataset']['title']}»"

        if result['title']:
            dataset_title += f"<div>({result['title']})</div>"
    else:
        dataset_title = result['title']

    match result['resultStatus']:
        case 'NO-HIT-GREEN':
            return f'Området er utenfor {dataset_title}'
        case 'NO-HIT-YELLOW':
            return f'Området har ikke treff for {dataset_title}'
        case 'HIT-YELLOW':
            return f'Området har treff i {dataset_title}'
        case 'HIT-RED':
            return f'Området er i konflikt med {dataset_title}'
        case 'TIMEOUT':
            return f'Tidsavbrudd: {dataset_title}'
        case 'ERROR':
            return f'En feil har oppstått: {dataset_title}'
        case _:
            return ''


def get_distance(result: Dict) -> str:
    distance = result['distanceToObject']

    if distance >= 20_000:
        distance = 20_000
        return f'> {format_number(distance)} m'

    return f'{format_number(distance)} m'


def get_hit_area(result: Dict) -> str:
    rounded = round(result['hitArea'])

    return f'{format_number(rounded)} m²'


def get_hit_area_percent(result: Dict) -> str:
    percent = (result['hitArea'] / result['inputGeometryArea']) * 100
    rounded = round(percent, 2)

    return f'{format_number(rounded)} %'


def get_input_area(result: Dict) -> str:
    rounded = round(result['inputGeometryArea'])

    return f'{format_number(rounded)} m²'


def format_number(value: int | float) -> str:
    return babel.numbers.format_decimal(value, locale='nb_NO')



import logging
from http import HTTPStatus
from pydantic import HttpUrl

_LOGGER = logging.getLogger(__name__)


def log_error_response(url: HttpUrl, status_code: int) -> None:
    try:
        status_txt = HTTPStatus(status_code).phrase
    except ValueError:
        status_txt = 'Unknown status code'

    err = f'Error: {url}: {status_txt} ({status_code})'

    _LOGGER.error(err)


__all__ = ['log_error_response']


import logging
import os
from urllib.parse import urlparse
from typing import Dict, Tuple, Union
from pydantic import HttpUrl, FileUrl
from osgeo import ogr
import aiohttp
import asyncio
from async_lru import alru_cache
from .gdal import query_gdal

_LOGGER = logging.getLogger(__name__)
_CACHE_TTL = 86400


async def query_geojson(url: Union[HttpUrl, FileUrl], filter: str, geometry: ogr.Geometry, epsg: int, timeout: int = 30) -> Dict:
    geojson = await _get_geojson(url, timeout)

    if not geojson:
        return None

    return query_gdal('GeoJSON', geojson, filter, geometry, epsg)


@alru_cache(maxsize=32, ttl=_CACHE_TTL)
async def _get_geojson(url: Union[HttpUrl, FileUrl], timeout) -> str:
    if url.scheme == 'file':
        geojson = _load_geojson(url)
    else:
        _, geojson = await _fetch_geojson(url, timeout)

    return geojson


async def _fetch_geojson(url: HttpUrl, timeout) -> Tuple[int, str]:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url, timeout=timeout) as response:
                if response.status != 200:
                    return response.status, None

                json_str = await response.text()

                return 200, json_str
    except asyncio.TimeoutError:
        return 408, None
    except Exception as err:
        _LOGGER.error(err)
        return 500, None


def _load_geojson(file_uri: FileUrl) -> str:
    path = _file_uri_to_path(file_uri)

    try:
        with open(path) as file:
            return file.read()
    except:
        return None


def _file_uri_to_path(file_uri: FileUrl) -> str:
    parsed = urlparse(str(file_uri))

    return os.path.abspath(os.path.join(parsed.netloc, parsed.path))





from typing import List, Dict
from .quality_measurement import QualityMeasurement


class CoverageQualityResponse():
    def __init__(self, quality_measurements: List[QualityMeasurement], warning_text: str, has_coverage: bool, is_relevant: bool, data: List[Dict]):
        self.quality_measurements = quality_measurements
        self.warning_text = warning_text
        self.has_coverage = has_coverage
        self.is_relevant = is_relevant
        self.data = data



import time
import logging
import traceback
from typing import List, Dict, Tuple
from uuid import UUID, uuid4
import asyncio
from socketio import SimpleClient
from osgeo import ogr
from pydash import kebab_case
from .dataset import get_config_ids, get_dataset_type
from .fact_sheet import create_fact_sheet
from .municipality import get_municipality
from .config import get_dataset_config
from .map_image import generate_map_images
from .report import create_pdf
from .blob_storage import create_container, upload_binary
from ..utils.helpers.geometry import create_input_geometry, get_epsg
from ..models.config import DatasetConfig
from ..models import Analysis, ArcGisAnalysis, OgcApiAnalysis, WfsAnalysis, EmptyAnalysis, AnalysisResponse, ResultStatus
from ..utils.constants import DEFAULT_EPSG
from ..utils.correlation_id_middleware import get_correlation_id

_LOGGER = logging.getLogger(__name__)


async def run(data: Dict, sio_client: SimpleClient) -> AnalysisResponse:
    geo_json = data.get('inputGeometry')
    geometry = create_input_geometry(geo_json)
    orig_epsg = get_epsg(geo_json)
    buffer = data.get('requestedBuffer', 0)
    context = data.get('context') or ''
    include_guidance = data.get('includeGuidance', False)
    include_quality_measurement = data.get('includeQualityMeasurement', False)
    include_facts = data.get('includeFacts', True)
    municipality_number, municipality_name = await get_municipality(geometry, DEFAULT_EPSG)

    datasets = await get_config_ids(data, municipality_number)
    correlation_id = get_correlation_id()

    if datasets and correlation_id and sio_client:
        to_analyze = {key: value for (
            key, value) in datasets.items() if value == True}
        sio_client.emit('datasets_counted_api', {'count': len(
            to_analyze), 'recipient': correlation_id})

    tasks: List[asyncio.Task] = []

    async with asyncio.TaskGroup() as tg:
        for config_id, should_analyze in datasets.items():
            task = tg.create_task(_run_analysis(
                config_id, should_analyze, geometry, DEFAULT_EPSG, orig_epsg, buffer,
                context, include_guidance, include_quality_measurement, sio_client))
            tasks.append(task)

    fact_sheet = None

    if include_facts:
        if correlation_id and sio_client:
            sio_client.emit('create_fact_sheet_api', {
                            'recipient': correlation_id})

        fact_sheet = await create_fact_sheet(geometry, orig_epsg, buffer)

    response = AnalysisResponse.create(
        geo_json, geometry, DEFAULT_EPSG, orig_epsg, buffer, fact_sheet, municipality_number, municipality_name)

    for task in tasks:
        response.result_list.append(task.result())

    analyses_with_map_image = [
        analysis for analysis in response.result_list if analysis.raster_result_map]

    if correlation_id and sio_client:
        sio_client.emit('create_map_images_api', {'recipient': correlation_id})

    map_images = generate_map_images(analyses_with_map_image, fact_sheet)

    container_name = str(uuid4())

    await _upload_images(response, map_images, container_name)

    if correlation_id and sio_client:
        sio_client.emit('create_report_api', {'recipient': correlation_id})

    report = create_pdf(response)

    response.report = await _upload_report(report, container_name)
    
    return response.to_dict()


async def _run_analysis(config_id: UUID, should_analyze: bool, geometry: ogr.Geometry, epsg: int, orig_epsg: int, buffer: int,
                        context: str, include_guidance: bool, include_quality_measurement: bool, sio_client: SimpleClient) -> Analysis:
    config = get_dataset_config(config_id)

    if config is None:
        return None

    if not should_analyze:
        analysis = EmptyAnalysis(
            config.config_id, config, ResultStatus.NOT_RELEVANT)
        await analysis.run()
        return analysis

    start = time.time()
    correlation_id = get_correlation_id()

    analysis = _create_analysis(
        config_id, config, geometry, epsg, orig_epsg, buffer)

    try:
        await analysis.run(context, include_guidance, include_quality_measurement)
    except Exception:
        err = traceback.format_exc()
        _LOGGER.error(err)
        await analysis.set_default_data()
        analysis.result_status = ResultStatus.ERROR

    end = time.time()

    # autopep8: off
    _LOGGER.info(f'Dataset analyzed: {config_id} - {config.name}: {round(end - start, 2)} sec.')
    # autopep8: on

    if correlation_id and sio_client:
        sio_client.emit('dataset_analyzed_api', {
            'dataset': str(config_id), 'recipient': correlation_id})

    return analysis


def _create_analysis(config_id: UUID, config: DatasetConfig, geometry: ogr.Geometry, epsg: int, orig_epsg: int, buffer: int) -> Analysis:
    dataset_type = get_dataset_type(config)

    match dataset_type:
        case 'arcgis':
            return ArcGisAnalysis(config_id, config, geometry, epsg, orig_epsg, buffer)
        case 'ogc_api':
            return OgcApiAnalysis(config_id, config, geometry, epsg, orig_epsg, buffer)
        case 'wfs':
            return WfsAnalysis(config_id, config, geometry, epsg, orig_epsg, buffer)
        case _:
            return None


async def _upload_images(response: AnalysisResponse, map_images: List[Tuple[str, str, bytes | None]], container_name: str) -> None:
    filtered = [map_image for map_image in map_images if map_image[2]]

    if not filtered:
        return

    await create_container(container_name)

    tasks: List[asyncio.Task[str]] = []

    async with asyncio.TaskGroup() as tg:
        for id, name, data in filtered:
            blob_name = f'{kebab_case(name)}.png'
            task = tg.create_task(upload_binary(
                data, container_name, blob_name, 'image/png'), name=id)
            tasks.append(task)

    for task in tasks:
        task_name = task.get_name()
        
        if task_name == 'omraade':
            response.fact_sheet.raster_result_image = task.result()
            continue

        analysis = _find_analysis(response.result_list, task_name)

        if analysis:
            analysis.raster_result_image = task.result()


async def _upload_report(report: bytes, container_name: str) -> str:
    await create_container(container_name)
    pdf_url = await upload_binary(report, container_name, 'rapport.pdf', 'application/pdf')

    return pdf_url


def _find_analysis(analyses: List[Analysis], config_id: str) -> Analysis:
    return next((analysis for analysis in analyses if str(analysis.config_id) == config_id), None)


__all__ = ['run']



import importlib.util
import importlib
import sys
import inspect
import logging
from io import BytesIO
from pathlib import Path
from types import ModuleType
from typing import List, Tuple, Dict, Any
from jinja2 import Environment, FileSystemLoader
from weasyprint import HTML
from ..models.analysis_response import AnalysisResponse
from ..utils.constants import PDF_TEMPLATES_DIR

logging.getLogger('weasyprint').setLevel(logging.ERROR)
logging.getLogger('fontTools.subset').setLevel(logging.ERROR)
logging.getLogger('fontTools.ttLib.ttFont').setLevel(logging.ERROR)
logging.getLogger('fontTools.varlib.mutator').setLevel(logging.ERROR)


def create_pdf(analysis_response: AnalysisResponse) -> bytes:
    templates_path = _get_templates_path()
    output = _render_html(templates_path, analysis_response.to_dict())

    buffer = BytesIO()
    html = HTML(string=output, base_url=templates_path)
    html.write_pdf(buffer)

    return buffer.getvalue()


def _render_html(templates_path: str, data: Dict) -> str:
    env = Environment(loader=FileSystemLoader(templates_path))
    functions = _get_helper_functions(f'{templates_path}/functions')

    for name, func in functions:
        env.globals[name] = func

    filters = _get_helper_functions(f'{templates_path}/filters')

    for name, func in filters:
        env.filters[name] = func

    template = env.get_template('index.jinja')

    return template.render(data=data)


def _get_helper_functions(path: str) -> List[Tuple[str, Any]]:
    helpers_dir = Path(path)

    if not helpers_dir.exists():
        return []

    python_files = list(helpers_dir.glob('*.py'))
    functions_list: List[Tuple[str, Any]] = []

    for file in python_files:
        module_name = file.name.replace('.py', '')
        module = _import_module_from_path(module_name, file)
        members = inspect.getmembers(module)

        functions = [member for member in members if inspect.isfunction(
            member[1]) and not member[0].startswith('_')]

        functions_list.extend(functions)

    return functions_list


def _import_module_from_path(module_name: str, file_path: str) -> ModuleType:
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    return module


def _get_templates_path() -> str:
    if PDF_TEMPLATES_DIR:
        return PDF_TEMPLATES_DIR

    return f'{Path(__file__).parents[1]}/pdf_templates'


__all__ = ['create_pdf']



import logging
import json
from typing import Tuple, Dict, Any
import aiohttp
import asyncio
from pydantic import HttpUrl
from cql2 import Expr
from osgeo import ogr
from . import log_error_response
from ..utils.helpers.geometry import geometry_to_wkt, transform_geometry
from ..utils.constants import WGS84_EPSG

_LOGGER = logging.getLogger(__name__)


async def query_ogc_api(base_url: HttpUrl, layer: str, geom_field: str, geometry: ogr.Geometry, filter: str, epsg: int, out_epsg: int = 4326, timeout: int = 30) -> Tuple[int, Dict]:
    has_search = await _has_search_endpoint(base_url)

    if has_search:
        return await _query_ogc_api_post(base_url, layer, geom_field, geometry, filter, epsg, out_epsg, timeout)

    return await _query_ogc_api_get(base_url, layer, geom_field, geometry, filter, epsg, out_epsg, timeout)


async def _query_ogc_api_get(base_url: HttpUrl, layer: str, geom_field: str, geometry: ogr.Geometry, filter: str, epsg: int, out_epsg: int = 4326, timeout: int = 30) -> Tuple[int, Dict]:
    url = _create_request_url(
        base_url, layer, geom_field, geometry, filter, epsg, out_epsg)

    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url, timeout=timeout) as response:
                if response.status != 200:
                    log_error_response(url, response.status)
                    return response.status, None

                return 200, await response.json()
    except asyncio.TimeoutError:
        return 408, None
    except Exception as err:
        _LOGGER.error(err)
        return 500, None


async def _query_ogc_api_post(base_url: HttpUrl, layer: str, geom_field: str, geometry: ogr.Geometry, filter: str, epsg: int, out_epsg: int = 4326, timeout: int = 30) -> Tuple[int, Dict]:
    request_body = _create_request_body(
        layer, geom_field, geometry, filter, epsg, out_epsg)

    json_str = json.dumps(request_body)

    url = f'{base_url}/search?f=json'

    try:
        async with aiohttp.ClientSession() as session:
            async with session.post(url, data=json_str, timeout=timeout) as response:
                if response.status != 200:
                    return response.status, None

                return 200, await response.json()
    except asyncio.TimeoutError:
        return 408, None
    except Exception as err:
        _LOGGER.error(err)
        return 500, None


def _create_request_url(base_url: HttpUrl, layer: str, geom_field: str, geometry: ogr.Geometry, filter: str, epsg: int, out_epsg: int = 4326) -> str:
    wkt_str = geometry_to_wkt(geometry, epsg)

    # autopep8: off
    filter_crs = f'&filter-crs=http://www.opengis.net/def/crs/EPSG/0/{epsg}' if epsg != WGS84_EPSG else ''
    crs = f'&crs=http://www.opengis.net/def/crs/EPSG/0/{out_epsg}' if out_epsg != WGS84_EPSG else ''
    url = f'{base_url}/collections/{layer}/items?f=json&filter-lang=cql2-text{filter_crs}{crs}&filter=S_INTERSECTS({geom_field},{wkt_str})'

    if filter:
        url += f' AND {filter}'
    # autopep8: on

    return url


def _create_request_body(layer: str, geom_field: str, geometry: ogr.Geometry, filter: str, epsg: int, out_epsg: int) -> Dict[str, Any]:
    transd = transform_geometry(geometry, epsg, WGS84_EPSG)
    transd.SwapXY()

    wkt_str = geometry_to_wkt(transd, WGS84_EPSG)
    cql2_text = f'S_INTERSECTS({geom_field}, {wkt_str})'

    if filter:
        cql2_text += f' AND {filter}'

    expr = Expr(cql2_text)

    cql2_json = {
        'collections': [layer],
        'filter': expr.to_json()
    }

    if out_epsg != WGS84_EPSG:
        cql2_json['crs'] = f'http://www.opengis.net/def/crs/EPSG/0/{out_epsg}'

    cql2_json['collections'] = [layer]
    cql2_json['filter']

    return cql2_json


async def _has_search_endpoint(base_url: str) -> bool:
    url = f'{base_url}/search'

    try:
        async with aiohttp.ClientSession() as session:
            async with session.head(url) as response:
                return response.status < 400
    except Exception as err:
        _LOGGER.error(err)
        return False


__all__ = ['query_ogc_api']
